#include "m.h"
#include <cstdlib>
#include <cmath> //dla sqrt

#define ABS(a)   ((a)>0?(a):(-a))

//prywatne

template<class T>
T SquareM<T>::ch_part(int k)
{
	T Akk;
	//wybor czesciowy
	int maxi=k;
	for (int j=k+1; j<=n; j++)
		if (ABS(A(maxi,k))<ABS(A(j,k))) maxi=j;
	Akk=A(maxi,k);
	if (Akk==0) return 0;

	//zamiana wierszy
	for (int j=1; j<=n; j++) 
	{T pom=A(k,j); A(k,j)=A(maxi,j); A(maxi,j)=pom;}
	for (int j=1; j<=m; j++) 
	{T pom=B(k,j); B(k,j)=B(maxi,j); B(maxi,j)=pom;}

	return Akk;
}

template<class T>
T SquareM<T>::ch_full(int k)
{
	T Akk;
	//wybor pelny
	int maxp=k;
	int maxq=k;
	for (int j=k; j<=n; j++)
		for (int l=k; l<=n; l++)
			if (ABS(A(maxp,maxq))<ABS(A(j,l))) {maxp=j; maxq=l;}
	Akk=A(maxp,maxq);
	if (Akk==0) return 0;

	//zamiana wierszy
	for (int j=1; j<=n; j++)
	{T pom=A(k,j); A(k,j)=A(maxp,j); A(maxp,j)=pom;}
	for (int j=1; j<=m; j++)
	{T pom=B(k,j); B(k,j)=B(maxp,j); B(maxp,j)=pom;}

	//zamiana kolumn k<->maxq
	for (int j=1; j<=n; j++)
	{T pom=A(j,k); A(j,k)=A(j,maxq); A(j,maxq)=pom;}
	int pom=per[k-1];
	per[k-1]=per[maxq-1];
	per[maxq-1]=pom;

	return Akk;
}

//publiczne

template<class T>
void SquareM<T>::SetA(T *tab)
{
	memcpy(tA,tab,n*n*sizeof(T));
}

template<class T>
void SquareM<T>::SetB(T *tab)
{
	memcpy(tB,tab,n*m*sizeof(T));
}

template<class T>
istream &operator>>(istream &s,SquareM<T> &m)
{
	cout << "Macierz>\n";
	for (int i=1; i<=m.n; i++)
	{
		for (int j=1; j<=m.n; j++) s >> m.A(i,j);
	}
	cout << "Wektor>\n";
	for (int i=1; i<=m.n; i++)
	{
		for (int j=1; j<=m.m; j++) s >> m.B(i,j);
	}
	return s;
}

template<class T>
ostream &operator<<(ostream &s,SquareM<T> &m)
{
	for (int i=1; i<=m.n; i++)
	{
		s << "[ ";
		for (int j=1; j<=m.n; j++) s << m.A(i,j) << " ";
		s << "]";
		s << "[ ";
		for (int j=1; j<=m.m; j++) s << m.B(i,j) << " ";
		s << "]";
		s << endl;
	}
	return s;
}

//z czesciowym wyborem
//zwraca 0 gdy det(A)=0
template<class T>
int SquareM<T>::Gauss_partial()
{
	mode=null;
	if (per) delete []per; per=0;
	for (int k=1; k<=n-1; k++)
	{
		T Akk=ch_part(k);
		for (int i=k+1; i<=n; i++)
		{
			T l=A(i,k)/Akk;
			for (int j=k; j<=n; j++)
				A(i,j)=A(i,j)-l*A(k,j);

			for (int j=1; j<=m; j++)
				B(i,j)=B(i,j)-l*B(k,j);
		}
	}
	if (A(n,n)==0) return 0;
	mode=part;
	return 1;
}

//z pelnym wyborem
//zwraca 0 gdy det(A)=0
template<class T>
int SquareM<T>::Gauss_full()
{
	mode=null;
	per=new int[n];
	for (int i=0; i<n; i++) per[i]=i;
	for (int k=1; k<=n-1; k++)
	{
		T Akk=ch_full(k);

		for (int i=k+1; i<=n; i++)
		{
			T l=A(i,k)/Akk;
			for (int j=k; j<=n; j++)
				A(i,j)=A(i,j)-l*A(k,j);

			for (int j=1; j<=m; j++)
				B(i,j)=B(i,j)-l*B(k,j);
		}
	}
	if (A(n,n)==0) return 0;
	mode=full;
	return 1;
}

//Gaussa-Jordana z pelnym wyborem
template<class T>
int SquareM<T>::Gauss_Jordan()
{
	mode=null;
	per=new int[n];
	for (int i=0; i<n; i++) per[i]=i;
	for (int k=1; k<=n; k++)
	{
		T Akk=ch_full(k);

		//dzielimy wiersz k przez Akk
		for (int j=k; j<=n; j++)
			A(k,j)/=Akk;

		for (int j=1; j<=m; j++)
			B(k,j)/=Akk;

		for (int i=1; i<=n; i++)
		if (i!=k)
		{
			T l=A(i,k);
			for (int j=k; j<=n; j++)
				A(i,j)=A(i,j)-l*A(k,j);

			for (int j=1; j<=m; j++)
				B(i,j)=B(i,j)-l*B(k,j);
		}
	}
	if (A(n,n)==0) return 0;
	mode=jordan;
	return 1;
}

//Rozklad bezposredni macierzy metoda Doolittle'a
template<class T>
int SquareM<T>::LU1()
{
	mode=null;
	
	for (int k=1; k<=n; k++)
	{
		for (int j=k; j<=n; j++)
		{
			T sum=0;
			for (int p=1; p<=k-1; p++) sum+=A(k,p)*A(p,j);
			A(k,j)-=sum;
		}

		T Akk=A(k,k);
		if (Akk==0) return 0;

		for (int i=k+1; i<=n; i++)
		{
			T sum=0;
			for (int p=1; p<=k-1; p++) sum+=A(i,p)*A(p,k);
			A(i,k)-=sum;
			A(i,k)/=Akk;
		}
	}

	mode=lu1;
	return 1;
}

//Rozklad bezposredni macierzy metoda Crouta
template<class T>
int SquareM<T>::LU2()
{
	mode=null;
	
	for (int k=1; k<=n; k++)
	{
		for (int i=k; i<=n; i++)
		{
			T sum=0;
			for (int p=1; p<=k-1; p++) sum+=A(i,p)*A(p,k);
			A(i,k)-=sum;
		}

		T Akk=A(k,k);
		if (Akk==0) return 0;

		for (int j=k+1; j<=n; j++)
		{
			T sum=0;
			for (int p=1; p<=k-1; p++) sum+=A(k,p)*A(p,j);
			A(k,j)-=sum;
			A(k,j)/=Akk;
		}
	}

	mode=lu2;
	return 1;
}

template<class T>
int SquareM<T>::D3()
{
	T *alfa=new T[n];
	T *beta=new T[n];
	alfa[0]=A(1,1);
	for (int i=1; i<n; i++)
	{
		if (alfa[i-1]==0) goto err;
		beta[i]=A(i+1,i)/alfa[i-1];
		alfa[i]=A(i+1,i+1)-beta[i]*A(i,i+1);
	}
	T *y=new T[n];
	for (int j=1; j<=m; j++)
	{
		y[0]=B(1,j);
		for (int i=1; i<n; i++)
			y[i]=B(i+1,j)-beta[i]*y[i-1];

		//wynik bedzie zapisywany w B
		B(n,j)=y[n-1]/alfa[n-1];
		for(int i=n-2; i>=0; i--)
			B(i+1,j)=(y[i]-A(i+1,i+2)*B(i+2,j))/alfa[i];
	}

	mode=d3;
	delete []y;
	delete []beta;
	delete []alfa;

	return 1;
err:
	delete []beta;
	delete []alfa;
	return 0;
}

//Rozklad macierzy metoda Choleskiego-Banachiewicza
template<class T>
int SquareM<T>::CB()
{
	mode=null;

	for (int j=1; j<=n; j++)
	{
		T sum=0;
		for (int k=1; k<=j-1; k++) sum+=A(j,k)*A(j,k);
		if (A(j,j)-sum<0) return 0;
		A(j,j)=sqrt(A(j,j)-sum);

		sum=0;
		for (int i=j+1; i<=n; i++)
		{
			for (int k=1; k<=j-1; k++) sum+=A(i,k)*A(j,k);
			A(i,j)=(A(i,j)-sum)/A(j,j);
		}
	}

	mode=cb;
	return 1;
}

//metoda iteracyjna Richarssona
template<class T>
int SquareM<T>::Richardson(int k,T p)
{
	//w maciezy B beda zapisywane kolejne przyblizenia
	T *x=new T[n];
	T *sum=new T[n];
	for (int col=1; col<=m; col++)
	{
		for (int i=1; i<=n; i++) x[i-1]=1;
		for (int l=0; l<k; l++)
		{
			for (int i=1; i<=n; i++)
			{
				sum[i-1]=0;
				for (int j=1; j<=n; j++)
				{
					sum[i-1]+=A(i,j)*x[j-1];
				}
			}
			for (int i=1; i<=n; i++)
				x[i-1]+=p*(B(i,col)-sum[i-1]);
		}
		for (int i=1; i<=n; i++) B(i,col)=x[i-1];
	}
	delete []sum;
	delete []x;

	mode=iter;

	return 1;
}

//metoda iteracyjna Jacobiego
template<class T>
int SquareM<T>::Jacob(int k)
{
	//w maciezy B beda zapisywane kolejne przyblizenia
	T *x=new T[n];
	T *sum=new T[n];
	for (int col=1; col<=m; col++)
	{
		for (int i=1; i<=n; i++) x[i-1]=1;
		for (int l=0; l<k; l++)
		{
			for (int i=1; i<=n; i++)
			{
				sum[i-1]=0;
				for (int j=1; j<=i-1; j++)
				{
					sum[i-1]+=A(i,j)*x[j-1];
				}

				for (int j=i+1; j<=n; j++)
				{
					sum[i-1]+=A(i,j)*x[j-1];
				}
			}
			for (int i=1; i<=n; i++)
				if (A(i,i)!=0)
				x[i-1]=(B(i,col)-sum[i-1])/A(i,i);
				else
				return 0;
		}
		for (int i=1; i<=n; i++) B(i,col)=x[i-1];
	}
	delete []sum;
	delete []x;

	mode=iter;

	return 1;
}

//metoda iteracyjna Gaussa-Seidla
template<class T>
int SquareM<T>::GS(int k)
{
	//w maciezy B beda zapisywane kolejne przyblizenia
	T *x=new T[n];
	for (int col=1; col<=m; col++)
	{
		for (int i=1; i<=n; i++) x[i-1]=1;
		for (int l=0; l<k; l++)
		{
			for (int i=1; i<=n; i++)
			{
				T sum=0;
				for (int j=1; j<=i-1; j++)
				{
					sum+=A(i,j)*x[j-1];
				}

				for (int j=i+1; j<=n; j++)
				{
					sum+=A(i,j)*x[j-1];
				}

				if (A(i,i)!=0)
				x[i-1]=(B(i,col)-sum)/A(i,i);
				else
				return 0;
			}
				
		}
		for (int i=1; i<=n; i++) B(i,col)=x[i-1];
	}
	delete []x;

	mode=iter;

	return 1;
}

//metoda iteracyjna SOR
template<class T>
int SquareM<T>::SOR(int k,T w)
{
	//w maciezy B beda zapisywane kolejne przyblizenia
	T *x=new T[n];
	for (int col=1; col<=m; col++)
	{
		for (int i=1; i<=n; i++) x[i-1]=1;
		for (int l=0; l<k; l++)
		{
			for (int i=1; i<=n; i++)
			{
				T sum=0;
				for (int j=1; j<=i-1; j++)
				{
					sum+=A(i,j)*x[j-1];
				}

				for (int j=i; j<=n; j++)
				{
					sum+=A(i,j)*x[j-1];
				}

				if (A(i,i)!=0)
				x[i-1]=x[i-1]+w*(B(i,col)-sum)/A(i,i);
				else
				return 0;
			}
				
		}
		for (int i=1; i<=n; i++) B(i,col)=x[i-1];
	}
	delete []x;

	mode=iter;

	return 1;
}

template<class T>
void SquareM<T>::Results(ostream &s)
{
	if (mode==null) return;
	T *w=new T[n*m]; //tablica wynikow

	if (mode==part) //czesciowy wybor
	{
		
		for (int i=n; i>0; i--)
		{
			for (int j=1; j<=m; j++)
			{

				T sum=0;
				for (int k=i+1; k<=n; k++)
				{
					if (per)
					sum+=A(i,k)*w[per[j-1]];
					else
					sum+=A(i,k)*w[(j-1)+(k-1)*m];
				}

				if (per)
				w[per[i-1]]=(B(i,j)-sum)/A(i,i);
				else
				w[(j-1)+(i-1)*m]=(B(i,j)-sum)/A(i,i);

			}
			
		}
	}
	else
	if (mode==full) //pelny wybor
	{
		
		for (int i=n; i>0; i--)
		{
			for (int j=1; j<=m; j++)
			{

				T sum=0;
				for (int k=i+1; k<=n; k++)
				{
					sum+=A(i,k)*w[(j-1)+(per[k-1])*m];
				}

				w[(j-1)+(per[i-1])*m]=(B(i,j)-sum)/A(i,i);

			}
		}
	}
	else
	if (mode==jordan) //Gaussa-Jordana
	{
		
		for (int i=1; i<=n; i++)
		{
			for (int j=1; j<=m; j++)
			{
				w[(j-1)+(per[i-1])*m]=B(i,j);
			}
			
		}
	}
	else
	if ((mode==iter)||(mode==d3))
	{
		for (int i=1; i<=n; i++)
		{	
			s << "[ ";
			for (int j=1; j<=m; j++)
				s << B(i,j) << " ";
			s << "]" << endl;
		}
	}

	if (mode<lu1)
	for (int i=0; i<n; i++)
	{
		s << "[ ";
		for (int j=0; j<m; j++)
			s << w[j+i*m] << " ";
		s << "]" << endl;
	}

	delete []w;
}

template<class T>
void SquareM<T>::LU(ostream &s)
{
	if (mode==lu1)
	{
		//Doolittla
		s << "L\n";
		for (int i=1; i<=n; i++)
		{
			cout << "[ ";
			for (int j=1; j<=n; j++)
				if (j==i)
					s << "1 ";
				else
				if (j>i)
					s << "0 ";
				else
					s << A(i,j) << " ";

			cout << "]\n";
		}

		s << "U\n";
		for (int i=1; i<=n; i++)
		{
			s << "[ ";
			for (int j=1; j<=n; j++)
				if (j>=i)
					s << A(i,j) << " ";
				else 
					s << "0 ";

			s << "]\n";
		}
	}
	else
	if (mode==lu2)
	{
		//Crouta
		s << "L\n";
		for (int i=1; i<=n; i++)
		{
			cout << "[ ";
			for (int j=1; j<=n; j++)
				if (j>i)
					s << "0 ";
				else
					s << A(i,j) << " ";
			cout << "]\n";
		}

		s << "U\n";
		for (int i=1; i<=n; i++)
		{
			s << "[ ";
			for (int j=1; j<=n; j++)
				if (j==i)
					s << "1 ";
				else
				if (j<i)
					s << "0 ";
				else
					s << A(i,j) << " ";

			s << "]\n";
		}
	}
}